/********************************************************************************
*                                                                               *
*      F i l e   I n f o r m a t i o n   a n d   M a n i p u l a t i o n        *
*                                                                               *
*********************************************************************************
* Copyright (C) 2000 by Jeroen van der Zijp.   All Rights Reserved.             *
*********************************************************************************
* Contributed by: Sean Hubbell                                                  *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*********************************************************************************
* $Id: FXFile.cpp,v 1.37 2000/10/31 04:12:32 jeroen Exp $                       *
********************************************************************************/
#include "xincs.h"
#include "fxver.h"
#include "fxdefs.h"
#include "FXStream.h"
#include "FXString.h"
#include "FXFile.h"



/*
  Notes:
  - Windows flavors of some of these functions are not perfect yet.
*/


/*******************************************************************************/

// Get current working directory
FXString FXFile::getCurrentDirectory(){
  FXchar buffer[MAXPATHLEN];
#ifndef WIN32
  if(getcwd(buffer,MAXPATHLEN)) return FXString(buffer);
#else
  if(GetCurrentDirectory(MAXPATHLEN,buffer)) return FXString(buffer);
#endif
  return "";
  }


// Change current directory
FXbool FXFile::setCurrentDirectory(const FXString& path){
#ifdef WIN32
  return !path.empty() && SetCurrentDirectory(path.text());
#else
  return !path.empty() && chdir(path.text())==0;
#endif
  }


// Get current drive prefix "a:", if any
// This is the same method as used in VC++ CRT.
FXString FXFile::getCurrentDrive(){
#ifdef WIN32
  FXchar buffer[MAXPATHLEN];
  if(GetCurrentDirectory(MAXPATHLEN,buffer) && isalpha(buffer[0]) && buffer[1]==':'){ 
    buffer[0]=tolower(buffer[0]);
    return FXString(buffer,2);
    }
#endif
  return "";
  }


// Change current drive prefix "a:"
// This is the same method as used in VC++ CRT.
FXbool FXFile::setCurrentDrive(const FXString& prefix){
#ifdef WIN32
  FXchar buffer[3];
  if(!prefix.empty() && isalpha(prefix[0]) && prefix[1]==':'){
    buffer[0]=prefix[0];
    buffer[1]=':';
    buffer[2]='\0';
    return SetCurrentDirectory(buffer);
    }
  return FALSE;
#else
  return TRUE;
#endif
  }


// Get user's home directory
FXString FXFile::getHomeDirectory(){
  static const FXchar* home=NULL;
  if(!home){
#ifdef WIN32
    HKEY hKey;
    char    szEnv[MAXPATHLEN+5];     // "HOME="
    DWORD   dwSize=MAXPATHLEN;
    if(RegOpenKeyEx(HKEY_CURRENT_USER,"Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders",0,KEY_READ,&hKey)==ERROR_SUCCESS){
      strcpy(szEnv,"HOME=");

      // Change "Personal" to "Desktop" if you want...
      if(RegQueryValueEx(hKey,"Personal",NULL,NULL,(LPBYTE)szEnv+5,&dwSize)==ERROR_SUCCESS){
#ifdef _MSC_VER
        _putenv(szEnv);
#else
        putenv(szEnv);
#endif
        }
      RegCloseKey(hKey);
      }
#endif
    home=getenv("HOME");
    if(!home) home="";
    }
  return FXString(home);
  }


// Get executable path 
FXString FXFile::getExecPath(){
  return FXString(getenv("PATH"));
  }


// Return directory part of pathname, assuming full pathname
FXString FXFile::directory(const FXString& file){
  register FXint n,i;
  if(!file.empty()){
    i=0;
#ifdef WIN32
    if(isalpha(file[0]) && file[1]==':') i=2;
#endif
    if(file[i]==PATHSEP) i++;
    n=i;
    while(file[i]){
      if(file[i]==PATHSEP) n=i; 
      i++;
      }
    return FXString(file.text(),n);
    }
  return "";
  }


// Return name and extension part of pathname
FXString FXFile::name(const FXString& file){
  register FXint f,n;
  if(!file.empty()){
    n=0;
#ifdef WIN32
    if(isalpha(file[0]) && file[1]==':') n=2;
#endif
    f=n;
    while(file[n]){
      if(file[n++]==PATHSEP) f=n; 
      }
    return FXString(file.text()+f,n-f);
    }
  return "";
  }


// Return file title, i.e. document name only:
//
//  aa        -> aa
//  aa.bb     -> aa
//  aa.bb.cc  -> aa.bb
//  .aa       -> .aa
FXString FXFile::title(const FXString& file){
  register FXint f,e,b,i;
  if(!file.empty()){
    i=0;
#ifdef WIN32
    if(isalpha(file[0]) && file[1]==':') i=2;
#endif
    f=i;
    while(file[i]){
      if(file[i++]==PATHSEP) f=i; 
      }
    b=f;
    if(file[b]=='.') b++;     // Leading `.'
    e=i;
    while(b<i){
      if(file[--i]=='.'){ e=i; break; }
      }
    return FXString(file.text()+f,e-f);
    }
  return "";
  }


// Return extension, if there is one:
//
//  aa        -> ""
//  aa.bb     -> bb
//  aa.bb.cc  -> cc
//  .aa       -> ""
FXString FXFile::extension(const FXString& file){
  register FXint f,e,i,n;
  if(!file.empty()){
    n=0;
#ifdef WIN32
    if(isalpha(file[0]) && file[1]==':') n=2;
#endif
    f=n;
    while(file[n]){
      if(file[n++]==PATHSEP) f=n; 
      }
    if(file[f]=='.') f++;     // Leading `.'
    e=i=n;
    while(f<i){
      if(file[--i]=='.'){ e=i+1; break; }
      }
    return FXString(file.text()+e,n-e);
    }
  return "";
  }


// Return drive letter prefix "c:"
FXString FXFile::drive(const FXString& file){
#ifdef WIN32
  FXchar buffer[3];
  if(!file.empty() && isalpha(file[0]) && file[1]==':'){
    buffer[0]=tolower(file[0]);
    buffer[1]=':';
    buffer[2]='\0';
    return FXString(buffer,2);
    }
#endif
  return "";
  }


// Perform tilde expansion
FXString FXFile::expand(const FXString& file){
#ifndef WIN32
  FXint len=file.length();
  if(0<len && file[0]=='~'){
    struct passwd *pwd;
    FXint i=1;
    while(i<len && file[i]!=PATHSEP) i++;
    if(i==1){
      pwd=getpwuid(getuid());
      }
    else{
      pwd=getpwnam(file.mid(1,i-1).text());
      }
    if(pwd && pwd->pw_dir){
      return pwd->pw_dir+file.mid(i,len-i);
      }
    }
#else
  if(!file.empty() && file[0]=='~' && (file[1]==PATHSEP || file[1]=='\0')){
    return getHomeDirectory()+file.mid(1,MAXPATHLEN);
    }
#endif
  return file;
  }


// Simplify a file path; the path will remain relative if it was relative
FXString FXFile::simplify(const FXString& file){
  if(!file.empty()){
    FXString result=file;
    FXint beg,end,head,cur;
    head=0;
#ifndef WIN32
    if(result[0]==PATHSEP){ 
      head++; 
      }
    else{ 
      while(result[head]=='.' && result[head+1]=='.' && result[head+2]==PATHSEP) head+=3; 
      if(result[head]=='.' && result[head+1]=='.' && result[head+2]==0) head+=2;
      }
#else
    if(result[0]==PATHSEP && result[1]==PATHSEP){    // UNC
      head+=2; 
      }
    else if(isalpha(result[0]) && result[1]==':' && result[2]==PATHSEP){ 
      head+=3; 
      }
    else if(result[0]==PATHSEP){ 
      head++; 
      }
    else{ 
      while(result[head]=='.' && result[head+1]=='.' && result[head+2]==PATHSEP) head+=3; 
      if(result[head]=='.' && result[head+1]=='.' && result[head+2]==0) head+=2;
      }
#endif
    for(beg=cur=head; file[beg]; beg=end){
      while(file[beg]==PATHSEP) beg++;
      for(end=beg; file[end] && file[end]!=PATHSEP; end++);
      if(beg==end) break;
      if(beg+1==end && file[beg]=='.') continue;
      if(beg+2==end && file[beg]=='.' && file[beg+1]=='.'){
        while(head<cur && result[cur-1]!=PATHSEP) cur--;
        continue;
        }
      if(head<cur && result[cur-1]!=PATHSEP) result[cur++]=PATHSEP;
      while(beg<end){ result[cur++]=file[beg++]; }
      }
    while(head<cur && result[cur-1]==PATHSEP) cur--;
    result[cur]=0;
    return result;
    }
  return "";
  }


// Build absolute pathname
FXString FXFile::absolute(const FXString& file){
  FXString pathfile=FXFile::simplify(FXFile::expand(file));
#ifndef WIN32
  if(pathfile[0]==PATHSEP) return pathfile;
#else
  if(pathfile[0]==PATHSEP && pathfile[1]==PATHSEP) return pathfile;   // UNC
  if(isalpha(pathfile[0]) && pathfile[1]==':' && pathfile[2]==PATHSEP) return pathfile;
  if(pathfile[0]==PATHSEP) return getCurrentDrive()+pathfile;
#endif
  return FXFile::simplify(getCurrentDirectory()+PATHSEPSTRING+pathfile);
  }


// Build absolute pathname from parts
FXString FXFile::absolute(const FXString& base,const FXString& file){
  FXString pathfile=FXFile::simplify(FXFile::expand(file));
#ifndef WIN32
  if(pathfile[0]==PATHSEP) return pathfile;
#else
  if(pathfile[0]==PATHSEP && pathfile[1]==PATHSEP) return pathfile;   // UNC
  if(isalpha(pathfile[0]) && pathfile[1]==':' && pathfile[2]==PATHSEP) return pathfile;
  if(pathfile[0]==PATHSEP) return getCurrentDrive()+pathfile;
#endif
  return FXFile::simplify(absolute(base)+PATHSEPSTRING+pathfile);
  }
  

// Search pathlist for file
FXString FXFile::search(const FXString& pathlist,const FXString& file){
  FXString pathfile=FXFile::simplify(FXFile::expand(file));
  FXString path;
  FXint beg,end;
#ifndef WIN32
  if(pathfile[0]==PATHSEP){
    if(exists(pathfile)) return pathfile;
    return "";
    }
#else
  if(pathfile[0]==PATHSEP && pathfile[1]==PATHSEP){
    if(exists(pathfile)) return pathfile;   // UNC
    return "";
    }
  if(isalpha(pathfile[0]) && pathfile[1]==':' && pathfile[2]==PATHSEP){
    if(exists(pathfile)) return pathfile;
    return "";
    }
  if(pathfile[0]==PATHSEP){
    pathfile=getCurrentDrive()+pathfile;
    if(exists(pathfile)) return pathfile;
    return "";
    }
#endif
  for(beg=0; pathlist[beg]; beg=end){
    while(pathlist[beg]==PATHLISTSEP) beg++;
    for(end=beg; pathlist[end] && pathlist[end]!=PATHLISTSEP; end++);
    if(beg==end) break;
    path=absolute(pathlist.mid(beg,end-beg),pathfile);
    if(exists(path)) return path;
    }
  return "";
  }


// Up one level, given absolute path
FXString FXFile::upLevel(const FXString& file){
  if(!file.empty()){
    FXint beg=0;
    FXint end=file.length();
#ifndef WIN32
    if(file[0]==PATHSEP) beg++;
#else
    if(file[0]==PATHSEP) beg++;
    else if(file[0]==PATHSEP && file[1]==PATHSEP) beg+=2;   // UNC
    else if(isalpha(file[0]) && file[1]==':' && file[2]==PATHSEP) beg+=3;
#endif
    if(beg<end && file[end-1]==PATHSEP) end--;
    while(beg<end && file[--end]!=PATHSEP);
    return file.mid(0,end);
    }
  return file;
  }


// Check if file represents absolute pathname
FXbool FXFile::isAbsolute(const FXString& file){
#ifndef WIN32
  return !file.empty() && file[0]==PATHSEP;
#else
  return !file.empty() && ((file[0]==PATHSEP && file[1]==PATHSEP) || (isalpha(file[0]) && file[1]==':' && file[2]==PATHSEP));
#endif
  }


// Does string represent topmost directory
FXbool FXFile::isTopDirectory(const FXString& file){
#ifndef WIN32
  return !file.empty() && (file[0]==PATHSEP && file[1]=='\0');
#else
  return !file.empty() && (file[0]==PATHSEP && file[1]=='\0') || (isalpha(file[0]) && file[1]==':' && file[2]==PATHSEP && file[3]=='\0');
#endif
  }


// Check if string represents a file
FXbool FXFile::isFile(const FXString& file){
#ifndef WIN32
  struct stat status;
  return !file.empty() && (stat(file.text(),&status)==0) && S_ISREG(status.st_mode);
#else
  DWORD atts;
  return !file.empty() && ((atts=GetFileAttributes(file.text()))!=0xFFFFFFFF) && !(atts&FILE_ATTRIBUTE_DIRECTORY);
#endif
  }


// Check if string represents a link
FXbool FXFile::isLink(const FXString& file){
#ifndef WIN32
  struct stat status;
  return !file.empty() && (stat(file.text(),&status)==0) && S_ISLNK(status.st_mode);
#else
  return FALSE;
#endif
  }

// Check if string represents a directory
FXbool FXFile::isDirectory(const FXString& file){
#ifndef WIN32
  struct stat status;
  return !file.empty() && (stat(file.text(),&status)==0) && S_ISDIR(status.st_mode);
#else
  DWORD atts;
  return !file.empty() && ((atts=GetFileAttributes(file.text()))!=0xFFFFFFFF) && (atts&FILE_ATTRIBUTE_DIRECTORY);
#endif
  }


// Return true if file is readable (thanks to gehriger@linkcad.com)
FXbool FXFile::isReadable(const FXString& file){
  return !file.empty() && access(file.text(),R_OK)==0;
  }


// Return true if file is writable (thanks to gehriger@linkcad.com)
FXbool FXFile::isWritable(const FXString& file){
  return !file.empty() && access(file.text(),W_OK)==0;
  }


// Return true if file is executable (thanks to gehriger@linkcad.com)
FXbool FXFile::isExecutable(const FXString& file){
  return !file.empty() && access(file.text(),X_OK)==0;
  }
  

// Check if owner has full permissions
FXbool FXFile::isOwnerReadWriteExecute(const FXString& file){
#ifndef WIN32
  struct stat status;
  return !file.empty() && (stat(file.text(),&status)==0) && (status.st_mode&S_IRUSR) && (status.st_mode&S_IWUSR) && (status.st_mode&S_IXUSR);
#else
  return TRUE;
#endif
  }


// Check if owner can read 
FXbool FXFile::isOwnerReadable(const FXString& file){
#ifndef WIN32
  struct stat status;
  return !file.empty() && (stat(file.text(),&status)==0) && (status.st_mode&S_IRUSR);
#else
  DWORD atts;
  return !file.empty() && ((atts=GetFileAttributes(file.text()))!=0xFFFFFFFF);
#endif
  }


// Check if owner can write
FXbool FXFile::isOwnerWritable(const FXString& file){
#ifndef WIN32
  struct stat status;
  return !file.empty() && stat(file.text(),&status)==0 && (status.st_mode&S_IWUSR);
#else
  DWORD atts;
  return !file.empty() && ((atts=GetFileAttributes(file.text()))!=0xFFFFFFFF) && !(atts&FILE_ATTRIBUTE_READONLY);
#endif
  }


// Check if owner can execute
FXbool FXFile::isOwnerExecutable(const FXString& file){
#ifndef WIN32
  struct stat status;
  return !file.empty() && (stat(file.text(),&status)==0) && (status.st_mode&S_IXUSR);
#else
  return TRUE;
#endif
  }


// Check if group has full permissions
FXbool FXFile::isGroupReadWriteExecute(const FXString& file){
#ifndef WIN32
  struct stat status;
  return !file.empty() && (stat(file.text(),&status)==0) && (status.st_mode&S_IRGRP) && (status.st_mode&S_IWGRP) && (status.st_mode&S_IXGRP);
#else
  return TRUE;
#endif
  }


// Check if group can read
FXbool FXFile::isGroupReadable(const FXString& file){
#ifndef WIN32
  struct stat status;
  return !file.empty() && (stat(file.text(),&status)==0) && (status.st_mode&S_IRGRP);
#else
  DWORD atts;
  return !file.empty() && ((atts=GetFileAttributes(file.text()))!=0xFFFFFFFF);
#endif
  }


// Check if group can write
FXbool FXFile::isGroupWritable(const FXString& file){
#ifndef WIN32
  struct stat status;
  return !file.empty() && (stat(file.text(),&status)==0) && (status.st_mode&S_IWGRP);
#else
  DWORD atts;
  return !file.empty() && ((atts=GetFileAttributes(file.text()))!=0xFFFFFFFF) && !(atts&FILE_ATTRIBUTE_READONLY);
#endif
  }


// Check if group can execute
FXbool FXFile::isGroupExecutable(const FXString& file){
#ifndef WIN32
  struct stat status;
  return !file.empty() && (stat(file.text(),&status)==0) && (status.st_mode&S_IXGRP);
#else
  return TRUE;
#endif
  }


// Check if everybody has full permissions
FXbool FXFile::isOtherReadWriteExecute(const FXString& file){
#ifndef WIN32
  struct stat status;
  return !file.empty() && (stat(file.text(),&status)==0) && (status.st_mode&S_IROTH) && (status.st_mode&S_IWOTH) && (status.st_mode&S_IXOTH);
#else
  return TRUE;
#endif
  }


// Check if everybody can read
FXbool FXFile::isOtherReadable(const FXString& file){
#ifndef WIN32
  struct stat status;
  return !file.empty() && (stat(file.text(),&status)==0) && (status.st_mode&S_IROTH);
#else
  DWORD atts;
  return !file.empty() && ((atts=GetFileAttributes(file.text()))!=0xFFFFFFFF);
#endif
  }


// Check if everybody can write
FXbool FXFile::isOtherWritable(const FXString& file){
#ifndef WIN32
  struct stat status;
  return !file.empty() && (stat(file.text(),&status)==0) && (status.st_mode&S_IWOTH);
#else
  DWORD atts;
  return !file.empty() && ((atts=GetFileAttributes(file.text()))!=0xFFFFFFFF) && !(atts&FILE_ATTRIBUTE_READONLY);
#endif
  }


// Check if everybody can execute
FXbool FXFile::isOtherExecutable(const FXString& file){
#ifndef WIN32
  struct stat status;
  return !file.empty() && (stat(file.text(),&status)==0) && (status.st_mode&S_IXOTH);
#else
  return TRUE;
#endif
  }


// Return time file was last modified
long FXFile::modified(const FXString& file){
#ifndef WIN32
  struct stat status;
  return !file.empty() && (stat(file.text(),&status)==0) ? (long)status.st_mtime : 0L;  
#else
  struct stat status;
  return !file.empty() && (stat(file.text(),&status)==0) ? (long)status.st_mtime : 0L;  
#endif
  }


// Return time file was last accessed
long FXFile::accessed(const FXString& file){
#ifndef WIN32
  struct stat status;
  return !file.empty() && (stat(file.text(),&status)==0) ? (long)status.st_atime : 0L;  
#else
  struct stat status;
  return !file.empty() && (stat(file.text(),&status)==0) ? (long)status.st_atime : 0L;  
#endif
  }


// Return time when created
long FXFile::created(const FXString& file){
#ifndef WIN32
  struct stat status;
  return !file.empty() && (stat(file.text(),&status)==0) ? (long)status.st_ctime : 0L;  
#else
  struct stat status;
  return !file.empty() && (stat(file.text(),&status)==0) ? (long)status.st_ctime : 0L;  
#endif
  }


// Return time when touched
long FXFile::touched(const FXString& file){
#ifndef WIN32
  struct stat status;
  return !file.empty() && (stat(file.text(),&status)==0) ? (long)FXMAX3(status.st_ctime,status.st_atime,status.st_mtime) : 0L;  
#else
  struct stat status;
  return !file.empty() && (stat(file.text(),&status)==0) ? (long)FXMAX3(status.st_ctime,status.st_atime,status.st_mtime) : 0L;  
#endif
  }


// Get file info
FXbool FXFile::info(const FXString& file,struct stat& info){
#ifndef WIN32
  return !file.empty() && (stat(file.text(),&info)==0);  
#else
  return !file.empty() && (stat(file.text(),&info)==0);  
#endif
  }


// Get file size
unsigned long FXFile::size(const FXString& file){
#ifndef WIN32
  struct stat status;
  return !file.empty() && (stat(file.text(),&status)==0) ? (unsigned long)status.st_size : 0L;  
#else
  struct stat status;
  return !file.empty() && (stat(file.text(),&status)==0) ? (unsigned long)status.st_size : 0L;  
#endif
  }


FXbool FXFile::exists(const FXString& file){
#ifndef WIN32
  struct stat status;
  return !file.empty() && (stat(file.text(),&status)==0);   
#else
  return !file.empty() && (GetFileAttributes(file.text())!=0xFFFFFFFF);
#endif
  }


// Return true if files are identical
FXbool FXFile::identical(const FXString& file1,const FXString& file2){
  if(file1!=file2){
    FXbool same=FALSE;
#ifndef WIN32
    struct stat stat1,stat2;
    return !::lstat(file1.text(),&stat1) && !::lstat(file2.text(),&stat2) && stat1.st_ino==stat2.st_ino;
#else
    HANDLE hFile1;
    HANDLE hFile2;
    hFile1=CreateFile(file1.text(),GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
    if(hFile1!=INVALID_HANDLE_VALUE){
      hFile2=CreateFile(file2.text(),GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
      if(hFile2!=INVALID_HANDLE_VALUE){
        BY_HANDLE_FILE_INFORMATION info1;
        BY_HANDLE_FILE_INFORMATION info2;
        if(GetFileInformationByHandle(hFile1,&info1) && GetFileInformationByHandle(hFile2,&info2)){
          same=(info1.nFileIndexLow==info2.nFileIndexLow && info1.nFileIndexHigh==info2.nFileIndexHigh && info1.dwVolumeSerialNumber==info2.dwVolumeSerialNumber);
          }
        CloseHandle(hFile2);
        }
      CloseHandle(hFile1);
      }
    return same;
#endif
    }
  return TRUE;
  }


// Return file mode flags
FXuint FXFile::mode(const FXString& file){
#ifndef WIN32
  struct stat status;
  return !file.empty() && (stat(file.text(),&status)==0) ? status.st_mode : 0;  
#else
  struct stat status;
  return !file.empty() && (stat(file.text(),&status)==0) ? status.st_mode : 0;  
#endif
  }


// Change the mode flags for this file
FXbool FXFile::mode(const FXString& file,FXuint mode){
#ifndef WIN32
  return !file.empty() && chmod(file.text(),mode)==0;
#else
  return FALSE; // Unimplemented yet
#endif
  }

//   /// Create new, not yet existing, directory
//   static FXbool createDirectory(const FXString& path,FXuint mode);
//   
//   /// Remove directory
//   static FXbool removeDirectory(const FXString& path);
//   
//   /// Create a new file
//   static FXbool createFile(const FXString& file,FXuint mode);
// 
//   /// Remove file
//   static FXbool removeFile(const FXString& file);
// 
// 
// // Create directory
// FXbool FXFile::createDirectory(const FXString& path,FXuint mode){
//   if(!path.empty()){
// #ifndef WIN32
//     struct stat status;
//     if(stat(path.text(),&status)==0){
//       return S_ISDIR(status.st_mode);
//       }
//     return mkdir(path.text(),mode)==0;
// #else
//     return FALSE; // Unimplemented yet
// #endif
//     }
//   return FALSE;
//   }
// 
// 
// // Remove directory
// FXbool FXFile::removeDirectory(const FXString& path){
// #ifndef WIN32
//   return rmdir(path.text())==0;
// #else
//   return FALSE; // Unimplemented yet
// #endif
//   }
// 
// 
// // Create a new file
// FXbool FXFile::createFile(const FXString& file,FXuint mode){
// #ifndef WIN32
//   FXint fd=open(file.text(),O_CREAT|O_WRONLY|O_TRUNC,mode);
//   if(fd>=0){ close(fd); return TRUE; }
//   return FALSE;
// #else
//   return FALSE; // Unimplemented yet
// #endif
//   }
// 
// 
// // Remove file
// FXbool FXFile::removeFile(const FXString& file){
// #ifndef WIN32
//   return unlink(file.text())==0;
// #else
//   return FALSE; // Unimplemented yet
// #endif
//   }


#ifndef WIN32

// Write bytes
static long fullwrite(int fd,const unsigned char *ptr,long len){
  long nwritten,ntotalwritten=0;
  while(len>0){
    nwritten=write(fd,ptr,len);
    if(nwritten<0){
#ifdef EINTR
      if(errno==EINTR) continue;
#endif
      return -1;
      }
    ntotalwritten+=nwritten;
    ptr+=nwritten;
    len-=nwritten;
    }
  return ntotalwritten;
  }


// Copy ordinary file
static FXbool copyfile(const FXString& oldfile,const FXString& newfile){
  unsigned char buffer[4096];
  struct stat status;
  long nread,nwritten;
  int src,dst;
  FXbool ok=FALSE;
  if((src=open(oldfile.text(),O_RDONLY))>=0){
    if(stat(oldfile.text(),&status)==0){
      if((dst=open(newfile.text(),O_WRONLY|O_CREAT|O_TRUNC,status.st_mode))>=0){
        while(1){
          nread=read(src,buffer,sizeof(buffer));
          if(nread<0){
#ifdef EINTR
            if(errno==EINTR) continue;
#endif
            goto err;
            }
          if(nread==0) break;
          nwritten=fullwrite(dst,buffer,nread);
          if(nwritten<0) goto err;
          if(nwritten!=nread) goto err;
          }
        ok=TRUE;
err:    close(dst);
        }
      }
    close(src);
    }
  return ok;
  }
  

// To search visited inodes
struct inodelist {
  ino_t st_ino;
  inodelist *next;
  };

  
// Forward declararion
static FXbool copyrec(const FXString& oldfile,const FXString& newfile,FXbool overwrite,inodelist* inodes);
  

// Copy directory
static FXbool copydir(const FXString& oldfile,const FXString& newfile,FXbool overwrite,struct stat& parentstatus,inodelist* inodes){
  DIR *dirp;
  struct dirent *dp;
  struct stat status;
  inodelist *in,inode;
  FXString oldchild,newchild;

  // See if visited this inode already
  for(in=inodes; in; in=in->next){
    if(in->st_ino==parentstatus.st_ino) return TRUE;
    }

  // Try make directory, if none exists yet
  if(mkdir(newfile.text(),parentstatus.st_mode|S_IWUSR)!=0 && errno!=EEXIST) return FALSE;

  // Can we stat it
  if(lstat(newfile.text(),&status)!=0 || !S_ISDIR(status.st_mode)) return FALSE;

  // Try open directory to copy
  dirp=opendir(oldfile.text());
  if(!dirp) return FALSE;

  // Add this to the list
  inode.st_ino=status.st_ino;
  inode.next=inodes;

  // Copy stuff
  while((dp=readdir(dirp))!=NULL){
    if(dp->d_name[0]!='.' || (dp->d_name[1]!='\0' && (dp->d_name[1]!='.' || dp->d_name[2]!='\0'))){
      oldchild=oldfile;
      if(oldchild[oldchild.length()-1]!=PATHSEP) oldchild.append(PATHSEP);
      oldchild.append(dp->d_name);
      newchild=newfile;
      if(newchild[newchild.length()-1]!=PATHSEP) newchild.append(PATHSEP);
      newchild.append(dp->d_name);
      if(!copyrec(oldchild,newchild,overwrite,&inode)){
        closedir(dirp);
        return FALSE;
        }
      }
    }

  // Close directory
  closedir(dirp);

  // Success
  return TRUE;
  }




// Recursive copy
static FXbool copyrec(const FXString& oldfile,const FXString& newfile,FXbool overwrite,inodelist* inodes){
  struct stat status1,status2;

  // Old file or directory does not exist
  if(::lstat(oldfile.text(),&status1)!=0) return FALSE;

  // If target is not a directory, remove it if allowed
  if(::lstat(newfile.text(),&status2)==0){
    if(!S_ISDIR(status2.st_mode)){
      if(!overwrite) return FALSE;
      FXTRACE((100,"unlink(%s)\n",newfile.text()));
      if(::unlink(newfile.text())!=0) return FALSE;
      }
    }

  // Source is direcotory: copy recursively
  if(S_ISDIR(status1.st_mode)){
    return copydir(oldfile,newfile,overwrite,status1,inodes);
    }

  // Source is regular file: copy block by block
  if(S_ISREG(status1.st_mode)){
    FXTRACE((100,"copyfile(%s,%s)\n",oldfile.text(),newfile.text()));
    return copyfile(oldfile,newfile);
    }

  // Source is fifo: make a new one
  if(S_ISFIFO(status1.st_mode)){
    FXTRACE((100,"mkfifo(%s)\n",newfile.text()));
    return ::mkfifo(newfile.text(),status1.st_mode);
    }

  // Source is device: make a new one
  if(S_ISBLK(status1.st_mode) || S_ISCHR(status1.st_mode) || S_ISSOCK(status1.st_mode)){
    FXTRACE((100,"mknod(%s)\n",newfile.text()));
    return ::mknod(newfile.text(),status1.st_mode,status1.st_rdev)==0;
    }

  // Source is symbolic link: make a new one
  if(S_ISLNK(status1.st_mode)){
    FXString lnkfile=FXFile::symlink(oldfile);
    FXTRACE((100,"symlink(%s,%s)\n",lnkfile.text(),newfile.text()));
    return ::symlink(lnkfile.text(),newfile.text())==0;
    }

  // This shouldn't happen
  return FALSE;
  }


#else


// Forward declararion
static FXbool copyrec(const FXString& oldfile,const FXString& newfile,FXbool overwrite);


// Copy directory
static FXbool copydir(const FXString& oldfile,const FXString& newfile,FXbool overwrite){
  FXString oldchild,newchild;
  DWORD atts;
  WIN32_FIND_DATA ffData;
  HANDLE hFindFile;

  // Try make directory, if none exists yet
  if(CreateDirectory(newfile.text(),NULL)==0 && GetLastError()!=ERROR_FILE_EXISTS) return FALSE;

  // Can we stat it
  if((atts=GetFileAttributes(newfile.text()))==0xffffffff || !(atts&FILE_ATTRIBUTE_DIRECTORY)) return FALSE;

  // Try open directory to copy
  hFindFile=FindFirstFile((oldfile+PATHSEPSTRING+"*").text(),&ffData);
  if(hFindFile==INVALID_HANDLE_VALUE) return FALSE;

  // Copy stuff
  do{
    if(ffData.cFileName[0]!='.' && (ffData.cFileName[1]!='\0' && (ffData.cFileName[1]!='.' || ffData.cFileName[2]!='\0'))){
      oldchild=oldfile;
      if(oldchild[oldchild.length()-1]!=PATHSEP) oldchild.append(PATHSEP);
      oldchild.append(ffData.cFileName);
      newchild=newfile;
      if(newchild[newchild.length()-1]!=PATHSEP) newchild.append(PATHSEP);
      newchild.append(ffData.cFileName);
      if(!copyrec(oldchild,newchild,overwrite)){
        FindClose(hFindFile);
        return FALSE;
        }
      }
    }
  while(FindNextFile(hFindFile,&ffData));

  // Close directory
  FindClose(hFindFile);

  // Success
  return TRUE;
  }


// Recursive copy
static FXbool copyrec(const FXString& oldfile,const FXString& newfile,FXbool overwrite){
  DWORD atts1,atts2;

  // Old file or directory does not exist
  if((atts1=GetFileAttributes(oldfile.text()))==0xffffffff) return FALSE;

  // If target is not a directory, remove it if allowed
  if((atts2=GetFileAttributes(newfile.text()))!=0xffffffff){
    if(!(atts2&FILE_ATTRIBUTE_DIRECTORY)){
      if(!overwrite) return FALSE;
      FXTRACE((100,"DeleteFile(%s)\n",newfile.text()));
      if(DeleteFile(newfile.text())!=0) return FALSE;
      }
    }

  // Source is direcotory: copy recursively
  if(atts1&FILE_ATTRIBUTE_DIRECTORY){
    return copydir(oldfile,newfile,overwrite);
    }

  // Source is regular file: copy block by block
  if(!(atts1&FILE_ATTRIBUTE_DIRECTORY)){
    FXTRACE((100,"CopyFile(%s,%s)\n",oldfile.text(),newfile.text()));
    return CopyFile(oldfile.text(),newfile.text(),!overwrite);
    }

  // This shouldn't happen
  return FALSE;
  }


#endif

/*
fxTraceLevel=101;
FXbool ok=FXFile::copy(argv[1],argv[2],0);
//FXbool ok=FXFile::remove(argv[1]);
fxmessage("ok=%d\n",ok);
//fxmessage("symlink(%s)=%s\n",argv[1],FXFile::symlink(argv[1]).text());
exit(0);
*/


// Copy file
FXbool FXFile::copy(const FXString& oldfile,const FXString& newfile,FXbool overwrite){
  if(newfile!=oldfile){
#ifndef WIN32
    return copyrec(oldfile,newfile,overwrite,NULL);
#else
    return copyrec(oldfile,newfile,overwrite);      // No symlinks, so no need to check if directories are visited already
#endif
    }
  return FALSE;
  }


// Remove file or directory
FXbool FXFile::remove(const FXString& file){
#ifndef WIN32
  struct stat status;
  if(::lstat(file.text(),&status)==0){
    if(S_ISDIR(status.st_mode)){
      DIR *dirp=::opendir(file.text());
      if(dirp){
        struct dirent *dp;    
        FXString child;
        while((dp=::readdir(dirp))!=NULL){
          if(dp->d_name[0]!='.' || (dp->d_name[1]!='\0' && (dp->d_name[1]!='.' || dp->d_name[2]!='\0'))){
            child=file;
            if(child[child.length()-1]!=PATHSEP) child.append(PATHSEP);
            child.append(dp->d_name);
            if(!FXFile::remove(child)){
              ::closedir(dirp);
              return FALSE;
              }
            }
          }
        ::closedir(dirp);
        }
      FXTRACE((100,"rmdir(%s)\n",file.text()));
      return ::rmdir(file.text())==0;
      }
    else{
      FXTRACE((100,"unlink(%s)\n",file.text()));
      return ::unlink(file.text())==0;
      }
    }
  return FALSE;
#else
  DWORD atts;
  if((atts=GetFileAttributes(file.text()))!=0xffffffff){
    if(atts&FILE_ATTRIBUTE_DIRECTORY){
      WIN32_FIND_DATA ffData;
      HANDLE hFindFile;
      hFindFile=FindFirstFile((file+PATHSEPSTRING+"*").text(),&ffData); // FIXME we may want to formalize the "walk over directory" in a few API's here also...
      if(hFindFile!=INVALID_HANDLE_VALUE){
        FXString child;
        do{
          if(ffData.cFileName[0]!='.' && (ffData.cFileName[1]!='\0' && (ffData.cFileName[1]!='.' || ffData.cFileName[2]!='\0'))){
            child=file;
            if(child[child.length()-1]!=PATHSEP) child.append(PATHSEP);
            child.append(ffData.cFileName);
            if(!FXFile::remove(child)){
              FindClose(hFindFile);
              return FALSE;
              }
            }
          }
        while(FindNextFile(hFindFile,&ffData));
        FindClose(hFindFile);
        }
      FXTRACE((100,"RemoveDirectory(%s)\n",file.text()));
      return RemoveDirectory(file.text())!=0;
      }
    else{
      FXTRACE((100,"DeleteFile(%s)\n",file.text()));
      return DeleteFile(file.text())!=0;
      }
    }
  return FALSE;
#endif
  }


// Rename or move file, or copy and delete old if different file systems
FXbool FXFile::move(const FXString& oldfile,const FXString& newfile,FXbool overwrite){
  if(newfile!=oldfile){
#ifndef WIN32
    if(!FXFile::exists(oldfile)) return FALSE;
    if(FXFile::exists(newfile)){
      if(!overwrite) return FALSE;
      if(!FXFile::remove(newfile)) return FALSE;
      }
    FXTRACE((100,"rename(%s,%s)\n",oldfile.text(),newfile.text()));
    if(::rename(oldfile.text(),newfile.text())==0) return TRUE;
    if(errno!=EXDEV) return FALSE;
    if(FXFile::copy(oldfile,newfile)){
      return FXFile::remove(oldfile);
      }
#else
    if(!FXFile::exists(oldfile)) return FALSE;
    if(FXFile::exists(newfile)){
      if(!overwrite) return FALSE;
      if(!FXFile::remove(newfile)) return FALSE;
      }
    FXTRACE((100,"MoveFile(%s,%s)\n",oldfile.text(),newfile.text()));
    if(::MoveFile(oldfile.text(),newfile.text())!=0) return TRUE;
    if(GetLastError()!=ERROR_NOT_SAME_DEVICE) return FALSE;
    if(FXFile::copy(oldfile,newfile)){
      return FXFile::remove(oldfile);
      }
#endif
    }
  return FALSE;
  }


// Link file
FXbool FXFile::link(const FXString& oldfile,const FXString& newfile,FXbool overwrite){
  if(newfile!=oldfile){
#ifndef WIN32
    if(!FXFile::exists(oldfile)) return FALSE;
    if(FXFile::exists(newfile)){
      if(!overwrite) return FALSE;
      if(!FXFile::remove(newfile)) return FALSE;
      }
    FXTRACE((100,"link(%s,%s)\n",oldfile.text(),newfile.text()));
    return ::link(oldfile.text(),newfile.text())==0;
#else
#if (_WIN32_WINNT >=0x0500)   // WinNT, Win2K, and later
    if(!FXFile::exists(oldfile)) return FALSE;
    if(FXFile::exists(newfile)){
      if(!overwrite) return FALSE;
      if(!FXFile::remove(newfile)) return FALSE;
      }
    FXTRACE((100,"CreateHardLink(%s,%s)\n",oldfile.text(),newfile.text()));
    return ::CreateHardLink(newfile.text(),oldfile.text(),NULL)!=0;
#else
    return FALSE;
#endif
#endif
    }
  return FALSE;
  }


// Symbolic Link file
FXbool FXFile::symlink(const FXString& oldfile,const FXString& newfile,FXbool overwrite){
#ifndef WIN32
  if(newfile!=oldfile){
    if(!FXFile::exists(oldfile)) return FALSE;
    if(FXFile::exists(newfile)){
      if(!overwrite) return FALSE;
      if(!FXFile::remove(newfile)) return FALSE;
      }
    FXTRACE((100,"symlink(%s,%s)\n",oldfile.text(),newfile.text()));
    return ::symlink(oldfile.text(),newfile.text())==0;
    }
#endif
  return FALSE;
  }


// Read symbolic link
FXString FXFile::symlink(const FXString& file){
#ifndef WIN32
  FXchar lnk[MAXPATHLEN+1];
  FXint len=::readlink(file.text(),lnk,MAXPATHLEN);
  if(0<=len) return FXString(lnk,len);
#endif
  return FXString::null;
  }

