/********************************************************************************
*                                                                               *
*                      T e x t   R e p l a c e   D i a l o g                    *
*                                                                               *
*********************************************************************************
* Copyright (C) 2000 by Jeroen van der Zijp.   All Rights Reserved.             *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*********************************************************************************
* $Id: FXReplaceDialog.cpp,v 1.2 2000/07/20 05:10:33 jeroen Exp $               *
********************************************************************************/
#include "xincs.h"
#include "fxver.h"
#include "fxdefs.h"
#include "fxkeys.h"
#include "FXStream.h"
#include "FXString.h"
#include "FXObject.h"
#include "FXObjectList.h"
#include "FXDict.h"
#include "FXRegistry.h"
#include "FXAccelTable.h"
#include "FXApp.h"
#include "FXId.h"
#include "FXFont.h"
#include "FXDrawable.h"
#include "FXImage.h"
#include "FXIcon.h"
#include "FXGIFIcon.h"
#include "FXWindow.h"
#include "FXDataTarget.h"
#include "FXFrame.h"
#include "FXLabel.h"
#include "FXButton.h"
#include "FXCheckButton.h"
#include "FXRadioButton.h"
#include "FXMenuButton.h"
#include "FXComposite.h"
#include "FXPacker.h"
#include "FXShell.h"
#include "FXPopup.h"
#include "FXTopWindow.h"
#include "FXDialogBox.h"
#include "FXScrollbar.h"
#include "FXTextField.h"
#include "FXScrollArea.h"
#include "FXVerticalFrame.h"
#include "FXHorizontalFrame.h"
#include "FXSearchDialog.h"
#include "FXReplaceDialog.h"



/*
  To do:
*/


/*******************************************************************************/

// Object implementation
FXIMPLEMENT(FXReplaceDialog,FXDialogBox,NULL,0)


// File Open Dialog
FXReplaceDialog::FXReplaceDialog(FXWindow* owner,const FXString& name,FXuint opts,FXint x,FXint y,FXint w,FXint h):
  FXDialogBox(owner,name,opts|DECOR_TITLE|DECOR_BORDER|DECOR_RESIZE,x,y,w,h),
  matchtgt(searchmode),dirtgt(backwards){
  FXVerticalFrame *contents=new FXVerticalFrame(this,LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 10,10,4,4);
  new FXLabel(contents,"&Search for:",NULL,LAYOUT_LEFT);
  searchtext=new FXTextField(contents,25,NULL,0,FRAME_SUNKEN|FRAME_THICK|LAYOUT_FILL_X,0,0,0,0, 8,8,4,4);
  new FXLabel(contents,"Replace &with:",NULL,LAYOUT_LEFT);
  replacetext=new FXTextField(contents,25,NULL,0,FRAME_SUNKEN|FRAME_THICK|LAYOUT_FILL_X,0,0,0,0, 8,8,4,4);
  FXHorizontalFrame* options1=new FXHorizontalFrame(contents,LAYOUT_FILL_X,0,0,0,0, 0,0,0,0);
  new FXRadioButton(options1,"L&iteral",&matchtgt,FXDataTarget::ID_OPTION+SEARCHMODE_EXACT,ICON_BEFORE_TEXT|LAYOUT_CENTER_X);
  new FXRadioButton(options1,"C&ase insensitive literal",&matchtgt,FXDataTarget::ID_OPTION+SEARCHMODE_IGNORECASE,ICON_BEFORE_TEXT|LAYOUT_CENTER_X);
  new FXRadioButton(options1,"Regular e&xpression",&matchtgt,FXDataTarget::ID_OPTION+SEARCHMODE_REGEX,ICON_BEFORE_TEXT|LAYOUT_CENTER_X);
  new FXCheckButton(options1,"&Backward",&dirtgt,FXDataTarget::ID_VALUE,ICON_BEFORE_TEXT|LAYOUT_CENTER_X);
  FXHorizontalFrame *buttons=new FXHorizontalFrame(contents,LAYOUT_BOTTOM|LAYOUT_FILL_X|PACK_UNIFORM_WIDTH,0,0,0,0, 0,0,0,0);
  lastbutton=new FXButton(buttons,"<< &Last\tSearch last",NULL,NULL,0,FRAME_RAISED|FRAME_THICK|LAYOUT_CENTER_X,0,0,0,0,6,6);
  nextbutton=new FXButton(buttons,"&Next >>\tSearch next",NULL,NULL,0,FRAME_RAISED|FRAME_THICK|LAYOUT_CENTER_X,0,0,0,0,6,6);
  replacebutton=new FXButton(buttons,"&Replace",NULL,this,ID_ACCEPT,BUTTON_DEFAULT|FRAME_RAISED|FRAME_THICK|LAYOUT_CENTER_X,0,0,0,0,6,6);
  replaceallbutton=new FXButton(buttons,"Replace &All",NULL,this,ID_ACCEPT,BUTTON_DEFAULT|FRAME_RAISED|FRAME_THICK|LAYOUT_CENTER_X,0,0,0,0,6,6);
  new FXButton(buttons,"&Cancel",NULL,this,ID_CANCEL,FRAME_RAISED|FRAME_THICK|LAYOUT_CENTER_X,0,0,0,0,6,6);
  searchtext->setFocus();
  searchmode=SEARCHMODE_EXACT;
  backwards=FALSE;
  }


// Set text or pattern to search for
void FXReplaceDialog::setSearchText(const FXString& text){
  searchtext->setText(text);
  }


// Return text or pattern the user has entered
FXString FXReplaceDialog::getSearchText() const {
  return searchtext->getText();
  }


// Set text or pattern to search with
void FXReplaceDialog::setReplaceText(const FXString& text){
  replacetext->setText(text);
  }


// Return text or pattern the user has entered
FXString FXReplaceDialog::getReplaceText() const {
  return replacetext->getText();
  }


// Save data
void FXReplaceDialog::save(FXStream& store) const {
  FXDialogBox::save(store);
  store << searchtext;
  store << replacetext;
  store << nextbutton;
  store << lastbutton;
  store << replacebutton;
  store << replaceallbutton;
  store << searchmode;
  store << backwards;
  }


// Load data
void FXReplaceDialog::load(FXStream& store){ 
  FXDialogBox::load(store);
  store >> searchtext;
  store >> replacetext;
  store >> nextbutton;
  store >> lastbutton;
  store >> replacebutton;
  store >> replaceallbutton;
  store >> searchmode;
  store >> backwards;
  }


// Cleanup
FXReplaceDialog::~FXReplaceDialog(){
  searchtext=(FXTextField*)-1;
  replacetext=(FXTextField*)-1;
  nextbutton=(FXButton*)-1;
  lastbutton=(FXButton*)-1;
  replacebutton=(FXButton*)-1;
  replaceallbutton=(FXButton*)-1;
  }


