/********************************************************************************
*                                                                               *
*                         F O X   T e x t   E d i t o r                         *
*                                                                               *
*********************************************************************************
* Copyright (C) 1998,2000 by Jeroen van der Zijp.   All Rights Reserved.        *
*********************************************************************************
* $Id: TextEdit.cpp,v 1.33 2000/11/01 23:31:12 jeroen Exp $                     *
********************************************************************************/
#include "fx.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <signal.h>
#ifndef WIN32
#include <unistd.h>
#endif
#include "HelpWindow.h"
#include "TextCommand.h"
#include "TextEdit.h"
#include "icons.h"




/*******************************************************************************/

  
// Patterns
const FXchar *TextWindow::patterns[]={
  "All Files",          "*",
  "C++ Source Files",   "*.cpp",
  "C Source Files",     "*.c",
  "C++ Header Files",   "*.h",
  "Text Files",         "*.txt",
  "HTML Header Files",  "*.html",
  NULL
  };


/*******************************************************************************/
  
// Map
FXDEFMAP(TextWindow) TextWindowMap[]={
  FXMAPFUNC(SEL_COMMAND,    TextWindow::ID_ABOUT,        TextWindow::onCmdAbout),
  FXMAPFUNC(SEL_COMMAND,    TextWindow::ID_REVERT,       TextWindow::onCmdRevert),
  FXMAPFUNC(SEL_UPDATE,     TextWindow::ID_REVERT,       TextWindow::onUpdRevert),
  FXMAPFUNC(SEL_COMMAND,    TextWindow::ID_OPEN,         TextWindow::onCmdOpen),
  FXMAPFUNC(SEL_COMMAND,    TextWindow::ID_SAVE,         TextWindow::onCmdSave),
  FXMAPFUNC(SEL_UPDATE,     TextWindow::ID_SAVE,         TextWindow::onUpdSave),
  FXMAPFUNC(SEL_COMMAND,    TextWindow::ID_SAVEAS,       TextWindow::onCmdSaveAs),
  FXMAPFUNC(SEL_COMMAND,    TextWindow::ID_NEW,          TextWindow::onCmdNew),
  FXMAPFUNC(SEL_UPDATE,     TextWindow::ID_TITLE,        TextWindow::onUpdTitle),
  FXMAPFUNC(SEL_COMMAND,    TextWindow::ID_FONT,         TextWindow::onCmdFont),
  FXMAPFUNC(SEL_COMMAND,    TextWindow::ID_QUIT,         TextWindow::onCmdQuit),
  FXMAPFUNC(SEL_SIGNAL,     TextWindow::ID_QUIT,         TextWindow::onCmdQuit),
  FXMAPFUNC(SEL_CLOSE,      TextWindow::ID_TITLE,        TextWindow::onCmdQuit),
  FXMAPFUNC(SEL_COMMAND,    TextWindow::ID_PRINT,        TextWindow::onCmdPrint),
  FXMAPFUNC(SEL_COMMAND,    TextWindow::ID_TREELIST,     TextWindow::onCmdTreeList),
  FXMAPFUNC(SEL_COMMAND,    TextWindow::ID_BACKCOLOR,    TextWindow::onCmdBackColor),
  FXMAPFUNC(SEL_CHANGED,    TextWindow::ID_BACKCOLOR,    TextWindow::onCmdBackColor),
  FXMAPFUNC(SEL_UPDATE,     TextWindow::ID_BACKCOLOR,    TextWindow::onUpdBackColor),
  FXMAPFUNC(SEL_COMMAND,    TextWindow::ID_SELBACKCOLOR, TextWindow::onCmdSelBackColor),
  FXMAPFUNC(SEL_CHANGED,    TextWindow::ID_SELBACKCOLOR, TextWindow::onCmdSelBackColor),
  FXMAPFUNC(SEL_UPDATE,     TextWindow::ID_SELBACKCOLOR, TextWindow::onUpdSelBackColor),
  FXMAPFUNC(SEL_COMMAND,    TextWindow::ID_TEXTCOLOR,    TextWindow::onCmdTextColor),
  FXMAPFUNC(SEL_CHANGED,    TextWindow::ID_TEXTCOLOR,    TextWindow::onCmdTextColor),
  FXMAPFUNC(SEL_UPDATE,     TextWindow::ID_TEXTCOLOR,    TextWindow::onUpdTextColor),
  FXMAPFUNC(SEL_COMMAND,    TextWindow::ID_SELTEXTCOLOR, TextWindow::onCmdSelTextColor),
  FXMAPFUNC(SEL_CHANGED,    TextWindow::ID_SELTEXTCOLOR, TextWindow::onCmdSelTextColor),
  FXMAPFUNC(SEL_UPDATE,     TextWindow::ID_SELTEXTCOLOR, TextWindow::onUpdSelTextColor),
  FXMAPFUNC(SEL_COMMAND,    TextWindow::ID_LINECOLOR,    TextWindow::onCmdLineColor),
  FXMAPFUNC(SEL_CHANGED,    TextWindow::ID_LINECOLOR,    TextWindow::onCmdLineColor),
  FXMAPFUNC(SEL_UPDATE,     TextWindow::ID_LINECOLOR,    TextWindow::onUpdLineColor),
  FXMAPFUNC(SEL_COMMAND,    TextWindow::ID_RECENTFILE,   TextWindow::onCmdRecentFile),
  FXMAPFUNC(SEL_UPDATE,     TextWindow::ID_TOGGLE_WRAP,  TextWindow::onUpdWrap),
  FXMAPFUNC(SEL_COMMAND,    TextWindow::ID_TOGGLE_WRAP,  TextWindow::onCmdWrap),
  FXMAPFUNC(SEL_COMMAND,    TextWindow::ID_SAVE_SETTINGS,TextWindow::onCmdSaveSettings),
  FXMAPFUNC(SEL_INSERTED,   TextWindow::ID_TEXT,         TextWindow::onTextInserted),
  FXMAPFUNC(SEL_DELETED,    TextWindow::ID_TEXT,         TextWindow::onTextDeleted),
  FXMAPFUNC(SEL_COMMAND,    TextWindow::ID_SEARCH,       TextWindow::onCmdSearch),
  FXMAPFUNC(SEL_COMMAND,    TextWindow::ID_REPLACE,      TextWindow::onCmdReplace),
  FXMAPFUNC(SEL_UPDATE,     TextWindow::ID_FIXED_WRAP,   TextWindow::onUpdWrapFixed),
  FXMAPFUNC(SEL_COMMAND,    TextWindow::ID_FIXED_WRAP,   TextWindow::onCmdWrapFixed),
  FXMAPFUNC(SEL_DND_MOTION, TextWindow::ID_TEXT,         TextWindow::onEditDNDMotion),
  FXMAPFUNC(SEL_DND_DROP,   TextWindow::ID_TEXT,         TextWindow::onEditDNDDrop),
  FXMAPFUNC(SEL_UPDATE,     TextWindow::ID_STRIP_CR,     TextWindow::onUpdStripReturns),
  FXMAPFUNC(SEL_COMMAND,    TextWindow::ID_STRIP_CR,     TextWindow::onCmdStripReturns),
  };


// Object implementation
FXIMPLEMENT(TextWindow,FXMainWindow,TextWindowMap,ARRAYNUMBER(TextWindowMap))


// Make some windows
TextWindow::TextWindow(FXApp* a):FXMainWindow(a,"FOX Text Editor: - untitled",NULL,NULL,DECOR_ALL,0,0,850,600,0,0){

  // Make some icons
  bigicon=new FXGIFIcon(getApp(),big_gif);
  smallicon=new FXGIFIcon(getApp(),small_gif);
  newicon=new FXGIFIcon(getApp(),new_gif);
  openicon=new FXGIFIcon(getApp(),open_gif);
  saveicon=new FXGIFIcon(getApp(),save_gif);
  saveasicon=new FXGIFIcon(getApp(),saveas_gif,0,IMAGE_ALPHAGUESS);
  printicon=new FXGIFIcon(getApp(),print_gif);
  cuticon=new FXGIFIcon(getApp(),cut_gif);
  copyicon=new FXGIFIcon(getApp(),copy_gif);
  pasteicon=new FXGIFIcon(getApp(),paste_gif);
  deleteicon=new FXGIFIcon(getApp(),delete_gif);
  undoicon=new FXGIFIcon(getApp(),undo_gif);
  redoicon=new FXGIFIcon(getApp(),redo_gif);
  fontsicon=new FXGIFIcon(getApp(),fonts_gif);
  colorsicon=new FXGIFIcon(getApp(),colors_gif);
  helpicon=new FXGIFIcon(getApp(),help_gif);

  // Application icons
  setIcon(bigicon);
  setMiniIcon(smallicon);
  
  // Make main window; set myself as the target
  setTarget(this);
  setSelector(ID_TITLE);
  
  // Help window
  helpwindow=new HelpWindow(this);
  
  // Make color dialog
  FXColorDialog *colordlg=new FXColorDialog(this,"Color Dialog");
  
  // Make menu bar
  FXToolbarShell* dragshell1=new FXToolbarShell(this,FRAME_RAISED|FRAME_THICK);
  FXMenubar* menubar=new FXMenubar(this,dragshell1,LAYOUT_SIDE_TOP|LAYOUT_FILL_X);
  new FXToolbarGrip(menubar,this,ID_TOOLBARGRIP);
  
  // Tool bar
  FXToolbarShell* dragshell2=new FXToolbarShell(this,FRAME_RAISED|FRAME_THICK);
  FXToolbar* toolbar=new FXToolbar(this,dragshell2,LAYOUT_SIDE_TOP|LAYOUT_FILL_X|PACK_UNIFORM_WIDTH|PACK_UNIFORM_HEIGHT);
  new FXToolbarGrip(toolbar,this,ID_TOOLBARGRIP);
  
  // Status bar
  FXStatusbar* statusbar=new FXStatusbar(this,LAYOUT_SIDE_BOTTOM|LAYOUT_FILL_X|STATUSBAR_WITH_DRAGCORNER);
  
  // Info about the editor
  new FXButton(statusbar,"\tThe FOX Text Editor\tAbout the FOX Text Editor.",smallicon,this,ID_ABOUT,LAYOUT_FILL_Y|LAYOUT_RIGHT);
  
  // File menu
  filemenu=new FXMenuPane(this);
  new FXMenuTitle(menubar,"&File",NULL,filemenu);
    
  // Edit Menu
  editmenu=new FXMenuPane(this);
  new FXMenuTitle(menubar,"&Edit",NULL,editmenu);
  
  // Options Menu
  optionmenu=new FXMenuPane(this);
  new FXMenuTitle(menubar,"Options",NULL,optionmenu);
  
  // View menu
  viewmenu=new FXMenuPane(this);
  new FXMenuTitle(menubar,"&View",NULL,viewmenu);
    
  // Help menu
  helpmenu=new FXMenuPane(this);
  new FXMenuTitle(menubar,"&Help",NULL,helpmenu,LAYOUT_RIGHT);

  // Splitter
  FXSplitter* splitter=new FXSplitter(this,LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y|SPLITTER_TRACKING);
  
  // Sunken border for tree
  treebox=new FXVerticalFrame(splitter,LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);
  
  // Make tree
  FXHorizontalFrame* treeframe=new FXHorizontalFrame(treebox,FRAME_SUNKEN|FRAME_THICK|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);
  dirlist=new FXDirList(treeframe,0,this,ID_TREELIST,DIRLIST_SHOWFILES|TREELIST_BROWSESELECT|TREELIST_SHOWS_LINES|TREELIST_SHOWS_BOXES|LAYOUT_FILL_X|LAYOUT_FILL_Y);
  FXHorizontalFrame* filterframe=new FXHorizontalFrame(treebox,LAYOUT_FILL_X);
  new FXLabel(filterframe,"Filter:");
  filter=new FXTextField(filterframe,10,dirlist,FXDirList::ID_SET_PATTERN,JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK|LAYOUT_FILL_X);
  
  // Sunken border for text widget
  FXHorizontalFrame *textbox=new FXHorizontalFrame(splitter,FRAME_SUNKEN|FRAME_THICK|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);
  
  // Make editor window
  editor=new FXText(textbox,this,ID_TEXT,LAYOUT_FILL_X|LAYOUT_FILL_Y);
  
  // Show column number in status bar
  new FXTextField(statusbar,4,editor,FXText::ID_CURSOR_COLUMN,FRAME_SUNKEN|FRAME_THICK|LAYOUT_RIGHT|LAYOUT_CENTER_Y,0,0,0,0,1,1,0,0);

  // Caption before number  
  new FXLabel(statusbar,"Col:",NULL,LAYOUT_RIGHT|LAYOUT_CENTER_Y);
  
  // Show line number in status bar
  new FXTextField(statusbar,4,editor,FXText::ID_CURSOR_ROW,FRAME_SUNKEN|FRAME_THICK|LAYOUT_RIGHT|LAYOUT_CENTER_Y,0,0,0,0,1,1,0,0);

  // Caption before number  
  new FXLabel(statusbar,"Line:",NULL,LAYOUT_RIGHT|LAYOUT_CENTER_Y);
  
  // Text color
  FXColorWell* textwell=new FXColorWell(statusbar,FXRGB(0,0,0),this,ID_TEXTCOLOR,FRAME_SUNKEN|FRAME_THICK|LAYOUT_RIGHT|LAYOUT_CENTER_Y);
  FXColorWell* backwell=new FXColorWell(statusbar,FXRGB(0,0,0),this,ID_BACKCOLOR,FRAME_SUNKEN|FRAME_THICK|LAYOUT_RIGHT|LAYOUT_CENTER_Y);
  FXColorWell* seltextwell=new FXColorWell(statusbar,FXRGB(0,0,0),this,ID_SELTEXTCOLOR,FRAME_SUNKEN|FRAME_THICK|LAYOUT_RIGHT|LAYOUT_CENTER_Y);
  FXColorWell* selbackwell=new FXColorWell(statusbar,FXRGB(0,0,0),this,ID_SELBACKCOLOR,FRAME_SUNKEN|FRAME_THICK|LAYOUT_RIGHT|LAYOUT_CENTER_Y);
  FXColorWell* linewell=new FXColorWell(statusbar,FXRGB(0,0,0),this,ID_LINECOLOR,FRAME_SUNKEN|FRAME_THICK|LAYOUT_RIGHT|LAYOUT_CENTER_Y);
  
  textwell->setTipText("Text");
  backwell->setTipText("Background");
  seltextwell->setTipText("Selection");
  selbackwell->setTipText("Selection background");
  linewell->setTipText("Directory inheritance lines");
  
  textwell->setHelpText("Change text color.");
  backwell->setHelpText("Change background color.");
  seltextwell->setHelpText("Change selection color.");
  selbackwell->setHelpText("Change selection background color.");
  linewell->setHelpText("Change directory inheritance line color");
  
  // Toobar buttons: File manipulation
  new FXButton(toolbar,"&New\tNew\tCreate new document.",newicon,this,ID_NEW,ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT);
  new FXButton(toolbar,"&Open\tOpen\tOpen document file.",openicon,this,ID_OPEN,ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT);
  new FXButton(toolbar,"&Save\tSave\tSave document.",saveicon,this,ID_SAVE,ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT);
  new FXButton(toolbar,"S&ave as\tSave As\tSave document to another file.",saveasicon,this,ID_SAVEAS,ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT);
  
  // Toobar buttons: Print
  new FXFrame(toolbar,LAYOUT_TOP|LAYOUT_LEFT|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT,0,0,5,5);
  new FXButton(toolbar,"&Print\tPrint\tPrint document.",printicon,this,ID_PRINT,ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT);
  
  // Toobar buttons: Editing
  new FXFrame(toolbar,LAYOUT_TOP|LAYOUT_LEFT|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT,0,0,5,5);
  new FXButton(toolbar,"Cut\tCut\tCut selection to clipboard.",cuticon,editor,FXText::ID_CUT_SEL,ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT);
  new FXButton(toolbar,"Copy\tCopy\tCopy selection to clipboard.",copyicon,editor,FXText::ID_COPY_SEL,ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT);
  new FXButton(toolbar,"Paste\tPaste\tPaste clipboard.",pasteicon,editor,FXText::ID_PASTE_SEL,ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT);
  //new FXButton(toolbar,"Undo\tUndo\tUndo last change.",undoicon,&undolist,FXUndoList::ID_UNDO,ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT);
  //new FXButton(toolbar,"Redo\tRedo\tRedo last undo.",redoicon,&undolist,FXUndoList::ID_REDO,ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT);
  
  // Color
  new FXFrame(toolbar,LAYOUT_TOP|LAYOUT_LEFT|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT,0,0,5,5);
  new FXButton(toolbar,"&Colors\tColors\tDisplay color dialog.",colorsicon,colordlg,FXWindow::ID_SHOW,ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT);
  new FXButton(toolbar,"Fon&ts\tFonts\tDisplay font dialog.",fontsicon,this,ID_FONT,ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT);
  
  new FXButton(toolbar,"&Help\tHelp on editor\tDisplay help information.",helpicon,helpwindow,ID_SHOW,ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_RIGHT);
  
  // File Menu entries
  new FXMenuCommand(filemenu,"&Open...        \tCtl-O\tOpen document file.",openicon,this,ID_OPEN);
  new FXMenuCommand(filemenu,"&Revert...\t\tRevert to saved.",NULL,this,ID_REVERT);
  new FXMenuCommand(filemenu,"&New...\tCtl-N\tCreate new document.",newicon,this,ID_NEW);
  new FXMenuCommand(filemenu,"&Save\tCtl-S\tSave changes to file.",saveicon,this,ID_SAVE);
  new FXMenuCommand(filemenu,"Save &As...\t\tSave document to another file.",saveasicon,this,ID_SAVEAS);
  new FXMenuCommand(filemenu,"&Print...\tCtl-P\tPrint document.",printicon,this,ID_PRINT);
  new FXMenuCommand(filemenu,"&Editable\t\tDocument editable.",NULL,editor,FXText::ID_TOGGLE_EDITABLE);
  new FXMenuCommand(filemenu,"&Iconify...\t\tIconify editor.",NULL,this,FXTopWindow::ID_ICONIFY);
  
  // Recent file menu; this automatically hides if there are no files
  FXMenuSeparator* sep1=new FXMenuSeparator(filemenu);
  sep1->setTarget(&mrufiles);
  sep1->setSelector(FXRecentFiles::ID_ANYFILES);
  new FXMenuCommand(filemenu,NULL,NULL,&mrufiles,FXRecentFiles::ID_FILE_1);
  new FXMenuCommand(filemenu,NULL,NULL,&mrufiles,FXRecentFiles::ID_FILE_2);
  new FXMenuCommand(filemenu,NULL,NULL,&mrufiles,FXRecentFiles::ID_FILE_3);
  new FXMenuCommand(filemenu,NULL,NULL,&mrufiles,FXRecentFiles::ID_FILE_4);
  new FXMenuCommand(filemenu,NULL,NULL,&mrufiles,FXRecentFiles::ID_FILE_5);
  new FXMenuCommand(filemenu,NULL,NULL,&mrufiles,FXRecentFiles::ID_FILE_6);
  new FXMenuCommand(filemenu,NULL,NULL,&mrufiles,FXRecentFiles::ID_FILE_7);
  new FXMenuCommand(filemenu,NULL,NULL,&mrufiles,FXRecentFiles::ID_FILE_8);
  new FXMenuCommand(filemenu,NULL,NULL,&mrufiles,FXRecentFiles::ID_FILE_9);
  new FXMenuCommand(filemenu,NULL,NULL,&mrufiles,FXRecentFiles::ID_FILE_10);
  new FXMenuCommand(filemenu,"&Clear Recent Files",NULL,&mrufiles,FXRecentFiles::ID_CLEAR);
  FXMenuSeparator* sep2=new FXMenuSeparator(filemenu);
  sep2->setTarget(&mrufiles);
  sep2->setSelector(FXRecentFiles::ID_ANYFILES);
  new FXMenuCommand(filemenu,"&Quit\tCtl-Q",NULL,this,ID_QUIT,MENU_DEFAULT);
  
  // Edit Menu entries
  new FXMenuCommand(editmenu,"&Undo\tCtl-Z\tUndo last change.",undoicon,&undolist,FXUndoList::ID_UNDO);
  new FXMenuCommand(editmenu,"&Redo\tCtl-Y\tRedo last undo.",redoicon,&undolist,FXUndoList::ID_REDO);
  new FXMenuSeparator(editmenu);
  new FXMenuCommand(editmenu,"&Copy\tCtl-C\tCopy selection to clipboard.",copyicon,editor,FXText::ID_COPY_SEL);
  new FXMenuCommand(editmenu,"Cu&t\tCtl-X\tCut selection to clipboard.",cuticon,editor,FXText::ID_CUT_SEL);
  new FXMenuCommand(editmenu,"&Paste\tCtl-V\tPaste from clipboard.",pasteicon,editor,FXText::ID_PASTE_SEL);
  new FXMenuCommand(editmenu,"&Delete\t\tDelete selection.",deleteicon,editor,FXText::ID_DELETE_SEL);
  new FXMenuSeparator(editmenu);
  new FXMenuCommand(editmenu,"&Search...\tCtl-F\tSearch for a string.",NULL,this,TextWindow::ID_SEARCH);
  new FXMenuCommand(editmenu,"R&eplace...\tCtl-R\tSearch for a string.",NULL,this,TextWindow::ID_REPLACE);
  
  // Options menu
  new FXMenuCommand(optionmenu,"Font...\t\tChange text font.",fontsicon,this,ID_FONT);
  new FXMenuCommand(optionmenu,"Colors...\t\tChange text colors.",colorsicon,colordlg,FXWindow::ID_SHOW);
  new FXMenuCommand(optionmenu,"Word Wrap\t\tToggle word wrap.",NULL,this,ID_TOGGLE_WRAP);
  new FXMenuCommand(optionmenu,"Fixed Wrap Column\t\tUse fixed columns for word wrap.",NULL,this,ID_FIXED_WRAP);
  new FXMenuCommand(optionmenu,"Hidden files\t\tShow hidden files and directories.",NULL,dirlist,FXDirList::ID_TOGGLE_HIDDEN);
  new FXMenuCommand(optionmenu,"Overstrike\t\tToggle overstrike mode.",NULL,editor,FXText::ID_TOGGLE_OVERSTRIKE);
  new FXMenuCommand(optionmenu,"Strip Returns\t\tToggle strip carriage returns mode.",NULL,this,TextWindow::ID_STRIP_CR);
  new FXMenuCommand(optionmenu,"Save Settings...\t\tSave settings now.",NULL,this,TextWindow::ID_SAVE_SETTINGS);
  
  // View Menu entries
  new FXMenuCommand(viewmenu,"File Browser\t\tDisplay file list.",NULL,treebox,FXWindow::ID_TOGGLESHOWN);
  new FXMenuCommand(viewmenu,"Toolbar\t\tDisplay toolbar.",NULL,toolbar,FXWindow::ID_TOGGLESHOWN);
  new FXMenuCommand(viewmenu,"Status line\t\tDisplay status line.",NULL,statusbar,FXWindow::ID_TOGGLESHOWN);
  
  // Help Menu entries
  new FXMenuCommand(helpmenu,"&Help...\t\tDisplay help information.",helpicon,helpwindow,ID_SHOW,0);
  new FXMenuSeparator(helpmenu);
  new FXMenuCommand(helpmenu,"&About TextEdit...\t\tDisplay about panel.",smallicon,this,ID_ABOUT,0);
  
  // Make a tool tip
  new FXTooltip(getApp(),0);
  
  // Recent files
  mrufiles.setTarget(this);
  mrufiles.setSelector(ID_RECENTFILE);

  // Initialize file name
  filename="untitled";
  namechanged=FALSE;
  modified=TRUE;
  }
  

// Clean up the mess
TextWindow::~TextWindow(){
  delete helpwindow;
  delete filemenu;
  delete editmenu;
  delete optionmenu;
  delete viewmenu;
  delete helpmenu;
  delete bigicon;
  delete smallicon;
  delete newicon;
  delete openicon;
  delete saveicon;
  delete saveasicon;
  delete printicon;
  delete cuticon;
  delete copyicon;
  delete pasteicon;
  delete deleteicon;
  delete undoicon;
  delete redoicon;
  delete fontsicon;
  delete colorsicon;
  delete helpicon;
  }


/*******************************************************************************/

// Load file
FXbool TextWindow::loadFile(const FXString& file){
  struct stat info;
  FILE *fp;
  char *text;
  FXuint n,i,j;
  if(stat(file.text(),&info)!=0){
    FXMessageBox::error(this,MBOX_OK,"Error Loading File","Non-existing file: %s",file.text());
    return FALSE;
    }
  fp=fopen(file.text(),"r");
  if(!fp){
    FXMessageBox::error(this,MBOX_OK,"Error Loading File","Unable to open file: %s",file.text());
    return FALSE;
    }
  getApp()->beginWaitCursor();
  // Under WIN32, when the file is opened in text mode, fread() converts CR/LF pairs
  // to single LF characters; so the number of items returned by fread() will be less
  // than the file size reported by info.st_size.
  FXCALLOC(&text,FXchar,info.st_size+1);
  n=fread(text,1,info.st_size,fp);
  
  // Strip carriage returns
  if(stripcr){
    for(i=j=0; j<n; j++){ 
      if(text[j]!='\r') text[i++]=text[j]; 
      } 
    text[i]='\0'; 
    n=i;
    }
  editor->setText(text,n);
  editor->setModified(FALSE);
  FXFREE(&text);
  fclose(fp);
  getApp()->endWaitCursor();
  undolist.clear();
  return TRUE;
  }


// Save file
FXbool TextWindow::saveFile(const FXString& file){
  FILE *fp;
  char *text;
  FXuint n,len;
  fp=fopen(file.text(),"w");
  if(!fp){
    FXMessageBox::error(this,MBOX_OK,"Error Saving File","Unable to open file: %s",file.text());
    return FALSE;
    }
  getApp()->beginWaitCursor();
  len=editor->getLength();
  FXCALLOC(&text,FXchar,len);
  editor->getText(text,len);
  n=fwrite(text,1,len,fp);
  if(n!=len){
    FXMessageBox::error(this,MBOX_OK,"Error Saving File","Unable to save entire file: %s",file.text());
    }
  fclose(fp);
  FXFREE(&text);
  editor->setModified(FALSE);
  getApp()->endWaitCursor();
  return TRUE;
  }


/*******************************************************************************/


// Read settings from registry
void TextWindow::readRegistry(){
  FXColor textback,textfore,textselback,textselfore,linecolor,cursorcolor;
  FXint ww,hh,xx,yy,treewidth,hidetree,hiddenfiles,wrapping,wrapcols,fixedwrap,tabcols;
  const FXchar* delimiters;
  FXFontDesc fontdesc;
  FXFont *font;
  
  // Get colors
  textback=getApp()->reg().readColorEntry("SETTINGS","textbackground",FXRGB(255,255,255));
  textfore=getApp()->reg().readColorEntry("SETTINGS","textforeground",FXRGB(0,0,0));
  textselback=getApp()->reg().readColorEntry("SETTINGS","textselbackground",FXRGB(0,0,128));
  textselfore=getApp()->reg().readColorEntry("SETTINGS","textselforeground",FXRGB(255,255,255));
  linecolor=getApp()->reg().readColorEntry("SETTINGS","directorytreelines",FXRGB(0,0,0));
  cursorcolor=getApp()->reg().readColorEntry("SETTINGS","cursorcolor",FXRGB(255,0,0));
  
  // Delimiters
  delimiters=getApp()->reg().readStringEntry("SETTINGS","delimiters","~.,/\\`'!@#$%^&*()-=+{}|[]\":;<>?");
  
  // Font
  if(fxparsefontdesc(fontdesc,getApp()->reg().readStringEntry("SETTINGS","font","helvetica 90 700 1 1 0 2"))){
    font=new FXFont(getApp(),fontdesc);
    editor->setFont(font);
    }
  
  // Get size
  xx=getApp()->reg().readIntEntry("SETTINGS","x",5);
  yy=getApp()->reg().readIntEntry("SETTINGS","y",5);
  ww=getApp()->reg().readIntEntry("SETTINGS","width",600);
  hh=getApp()->reg().readIntEntry("SETTINGS","height",400);

  // Hidden files shown
  hiddenfiles=getApp()->reg().readIntEntry("SETTINGS","showhiddenfiles",FALSE);
  dirlist->showHiddenFiles(hiddenfiles);
  
  // Showing the tree?
  hidetree=getApp()->reg().readIntEntry("SETTINGS","hidetree",TRUE);

  // Width of tree
  treewidth=getApp()->reg().readIntEntry("SETTINGS","treewidth",100);

  // Word wrapping
  wrapping=getApp()->reg().readIntEntry("SETTINGS","wordwrap",0);
  wrapcols=getApp()->reg().readIntEntry("SETTINGS","wrapcols",80);
  fixedwrap=getApp()->reg().readIntEntry("SETTINGS","fixedwrap",1);
  
  // Tab columns
  tabcols=getApp()->reg().readIntEntry("SETTINGS","tabcols",8);

  // Strip returns
  stripcr=getApp()->reg().readIntEntry("SETTINGS","stripreturn",FALSE);
  
  // Change the colors
  editor->setTextColor(textfore);
  dirlist->setTextColor(textfore);
  editor->setBackColor(textback);
  dirlist->setBackColor(textback);
  editor->setSelBackColor(textselback);
  dirlist->setSelBackColor(textselback);
  editor->setSelTextColor(textselfore);
  dirlist->setSelTextColor(textselfore);
  dirlist->setLineColor(linecolor);
  editor->setCursorColor(cursorcolor);
  
  // Change delimiters
  FXCharset set(delimiters);
  editor->setDelimiters(set);

  // Hide tree if asked for
  if(hidetree) treebox->hide();
  
  // Set tree width
  treebox->setWidth(treewidth);
  
  // Open toward file
  dirlist->setCurrentFile(filename);
  
  // Wrap mode
  if(wrapping) 
    editor->setTextStyle(editor->getTextStyle()|TEXT_WORDWRAP);
  else
    editor->setTextStyle(editor->getTextStyle()&~TEXT_WORDWRAP);
    
  // Wrap fixed mode
  if(fixedwrap) 
    editor->setTextStyle(editor->getTextStyle()|TEXT_FIXEDWRAP);
  else
    editor->setTextStyle(editor->getTextStyle()&~TEXT_FIXEDWRAP);
  
  // Wrap and tab columns
  editor->setWrapColumns(wrapcols);
  editor->setTabColumns(tabcols);
  
  // Reposition window
  position(xx,yy,ww,hh);
  }


// Save settings to registry
void TextWindow::writeRegistry(){
  char fontspec[200];
  FXFontDesc fontdesc;
  
  // Write changed colors back to registry
  getApp()->reg().writeColorEntry("SETTINGS","textbackground",editor->getBackColor());
  getApp()->reg().writeColorEntry("SETTINGS","textforeground",editor->getTextColor());
  getApp()->reg().writeColorEntry("SETTINGS","textselbackground",editor->getSelBackColor());
  getApp()->reg().writeColorEntry("SETTINGS","textselforeground",editor->getSelTextColor());
  getApp()->reg().writeColorEntry("SETTINGS","directorytreelines",dirlist->getLineColor());
  getApp()->reg().writeColorEntry("SETTINGS","cursorcolor",editor->getCursorColor());
  
  // Delimiters
  FXString delimiters=editor->getDelimiters();
  getApp()->reg().writeStringEntry("SETTINGS","delimiters",delimiters.text());
  
  // Write new window size back to registry
  getApp()->reg().writeIntEntry("SETTINGS","x",getX());
  getApp()->reg().writeIntEntry("SETTINGS","y",getY());
  getApp()->reg().writeIntEntry("SETTINGS","width",getWidth());
  getApp()->reg().writeIntEntry("SETTINGS","height",getHeight());
  
  // Were showing hidden files
  getApp()->reg().writeIntEntry("SETTINGS","showhiddenfiles",dirlist->showHiddenFiles());
  
  // Was tree shown?
  getApp()->reg().writeIntEntry("SETTINGS","hidetree",!treebox->shown());
  
  // Width of tree
  getApp()->reg().writeIntEntry("SETTINGS","treewidth",treebox->getWidth());
  
  // Wrap mode
  getApp()->reg().writeIntEntry("SETTINGS","wordwrap",(editor->getTextStyle()&TEXT_WORDWRAP)!=0);
  getApp()->reg().writeIntEntry("SETTINGS","fixedwrap",(editor->getTextStyle()&TEXT_FIXEDWRAP)!=0);
  getApp()->reg().writeIntEntry("SETTINGS","wrapcols",editor->getWrapColumns());
  
  // Tab columns
  getApp()->reg().writeIntEntry("SETTINGS","tabcols",editor->getTabColumns());

  // Strip returns
  getApp()->reg().writeIntEntry("SETTINGS","stripreturn",stripcr);
  
  // Font
  editor->getFont()->getFontDesc(fontdesc);
  fxunparsefontdesc(fontspec,fontdesc);
  getApp()->reg().writeStringEntry("SETTINGS","font",fontspec);
  }


/*******************************************************************************/


// About box
long TextWindow::onCmdAbout(FXObject*,FXSelector,void*){
  FXMessageBox about(this,"About Text Edit","Text Editor Demonstrates the FOX Text Widget.\n\nUsing the FOX C++ GUI Library (http://www.cfdrc.com/FOX/fox.html)\n\nCopyright (C) 2000 Jeroen van der Zijp (jvz@cfdrc.com)",bigicon,MBOX_OK|DECOR_TITLE|DECOR_BORDER);
  about.execute();  
  return 1;
  }


// Change font
long TextWindow::onCmdFont(FXObject*,FXSelector,void*){
  FXFontDialog fontdlg(this,"Change Font",DECOR_BORDER|DECOR_TITLE);
  FXFont *oldfont=editor->getFont();
  FXFontDesc fontdesc;
  oldfont->getFontDesc(fontdesc);
  fontdlg.setFontSelection(fontdesc);
  if(fontdlg.execute()){
    fontdlg.getFontSelection(fontdesc);
    FXFont *newfont=new FXFont(getApp(),fontdesc);
    newfont->create();
    editor->setFont(newfont);
    if(oldfont!=getApp()->getNormalFont()){
      delete oldfont;
      }
    }
  editor->setFocus();
  return 1;
  }


// Save settings
long TextWindow::onCmdSaveSettings(FXObject*,FXSelector,void*){
  writeRegistry();
  getApp()->reg().write();
  return 1;
  }


// Toggle wrap mode
long TextWindow::onCmdWrap(FXObject*,FXSelector,void*){
  editor->setTextStyle(editor->getTextStyle()^TEXT_WORDWRAP);
  return 1;
  }


// Update toggle wrap mode
long TextWindow::onUpdWrap(FXObject* sender,FXSelector,void*){
  if(editor->getTextStyle()&TEXT_WORDWRAP)
    sender->handle(this,MKUINT(ID_CHECK,SEL_COMMAND),NULL);
  else
    sender->handle(this,MKUINT(ID_UNCHECK,SEL_COMMAND),NULL);
  return 1;
  }


// Toggle fixed wrap mode
long TextWindow::onCmdWrapFixed(FXObject*,FXSelector,void*){
  editor->setTextStyle(editor->getTextStyle()^TEXT_FIXEDWRAP);
  return 1;
  }


// Update toggle fixed wrap mode
long TextWindow::onUpdWrapFixed(FXObject* sender,FXSelector,void*){
  if(editor->getTextStyle()&TEXT_FIXEDWRAP)
    sender->handle(this,MKUINT(ID_CHECK,SEL_COMMAND),NULL);
  else
    sender->handle(this,MKUINT(ID_UNCHECK,SEL_COMMAND),NULL);
  return 1;
  }

// Toggle strip returns mode
long TextWindow::onCmdStripReturns(FXObject*,FXSelector,void*){
  stripcr=!stripcr;
  return 1;
  }


// Update toggle strip returns mode
long TextWindow::onUpdStripReturns(FXObject* sender,FXSelector,void*){
  if(stripcr)
    sender->handle(this,MKUINT(ID_CHECK,SEL_COMMAND),NULL);
  else
    sender->handle(this,MKUINT(ID_UNCHECK,SEL_COMMAND),NULL);
  return 1;
  }


// Revert to saved
long TextWindow::onCmdRevert(FXObject*,FXSelector,void*){
  loadFile(filename);
  return 1;
  }


// Revert Update
long TextWindow::onUpdRevert(FXObject* sender,FXSelector,void* ptr){
  if(modified!=editor->isModified()){
    modified=editor->isModified();
    FXuint msg=modified ? FXWindow::ID_ENABLE : FXWindow::ID_DISABLE;
    sender->handle(this,MKUINT(msg,SEL_COMMAND),ptr);
    }
  return 1;
  }


// Open
long TextWindow::onCmdOpen(FXObject*,FXSelector,void*){
  FXuint answer;
  if(editor->isModified()){
    answer=FXMessageBox::question(this,MBOX_YES_NO_CANCEL,"Unsaved Document","Save current document to file?");
    if(answer==MBOX_CLICKED_CANCEL) return 1;
    if(answer==MBOX_CLICKED_YES){
      if(!namechanged){
        FXFileDialog savedialog(this,"Save Document");
        savedialog.setFilename(filename);
        if(!savedialog.execute()) return 1;
        if(fxexists(savedialog.getFilename().text())){
          if(MBOX_CLICKED_NO==FXMessageBox::question(this,MBOX_YES_NO,"Overwrite Document","Overwrite existing document?")) return 1;
          }
        filename=savedialog.getFilename();
        }
      saveFile(filename);
      }
    }
  FXFileDialog open(this,"Open Document");
  open.setFilename(filename);
  open.setPatternList(patterns);
  if(open.execute()){
    filename=open.getFilename();
    dirlist->setCurrentFile(filename);
    mrufiles.appendFile(filename);
    loadFile(filename);
    namechanged=TRUE;
    }
  return 1;
  }


// Open recent file
long TextWindow::onCmdRecentFile(FXObject*,FXSelector,void* ptr){
  FXuint answer;
  if(editor->isModified()){
    answer=FXMessageBox::question(this,MBOX_YES_NO_CANCEL,"Unsaved Document","Save current document to file?");
    if(answer==MBOX_CLICKED_CANCEL) return 1;
    if(answer==MBOX_CLICKED_YES){
      if(!namechanged){
        FXFileDialog savedialog(this,"Save Document");
        savedialog.setFilename(filename);
        if(!savedialog.execute()) return 1;
        if(fxexists(savedialog.getFilename().text())){
          if(MBOX_CLICKED_NO==FXMessageBox::question(this,MBOX_YES_NO,"Overwrite Document","Overwrite existing document?")) return 1;
          }
        filename=savedialog.getFilename();
        }
      saveFile(filename);
      }
    }
  filename=(FXchar*)ptr;
  dirlist->setCurrentFile(filename);
  loadFile(filename);
  namechanged=TRUE;
  return 1;
  }


// Save
long TextWindow::onCmdSave(FXObject* sender,FXSelector sel,void* ptr){
  if(!namechanged) return onCmdSaveAs(sender,sel,ptr);
  saveFile(filename);
  return 1;
  }


// Save Update
long TextWindow::onUpdSave(FXObject* sender,FXSelector,void* ptr){
  FXuint msg=editor->isModified() ? FXWindow::ID_ENABLE : FXWindow::ID_DISABLE;
  sender->handle(this,MKUINT(msg,SEL_COMMAND),ptr);
  return 1;
  }


// Save As
long TextWindow::onCmdSaveAs(FXObject*,FXSelector,void*){
  FXFileDialog savedialog(this,"Save Document");
  savedialog.setFilename(filename);
  if(savedialog.execute()){
    if(fxexists(savedialog.getFilename().text())){
      if(MBOX_CLICKED_NO==FXMessageBox::question(this,MBOX_YES_NO,"Overwrite Document","Overwrite existing document?")) return 1;
      }
    filename=savedialog.getFilename();
    dirlist->setCurrentFile(filename);
    mrufiles.appendFile(filename);
    saveFile(filename);
    namechanged=TRUE;
    }
  return 1;
  }


// New
long TextWindow::onCmdNew(FXObject*,FXSelector,void*){
  FXuint answer;
  if(editor->isModified()){
    answer=FXMessageBox::question(this,MBOX_YES_NO_CANCEL,"Unsaved Document","Save current document to file?");
    if(answer==MBOX_CLICKED_CANCEL) return 1;
    if(answer==MBOX_CLICKED_YES){
      if(!namechanged){
        FXFileDialog savedialog(this,"Save Document");
        savedialog.setFilename(filename);
        if(!savedialog.execute()) return 1;
        if(fxexists(savedialog.getFilename().text())){
          if(MBOX_CLICKED_NO==FXMessageBox::question(this,MBOX_YES_NO,"Overwrite Document","Overwrite existing document?")) return 1;
          }
        filename=savedialog.getFilename();
        }
      saveFile(filename);
      }
    }
  filename="untitled";
  editor->setText(NULL,0);
  namechanged=FALSE;
  editor->setModified(FALSE);
  return 1;
  }


// Quit
long TextWindow::onCmdQuit(FXObject*,FXSelector,void*){
  FXuint answer;
  if(editor->isModified()){
    answer=FXMessageBox::question(this,MBOX_YES_NO_CANCEL,"Unsaved Document","Save current document to file?");
    if(answer==MBOX_CLICKED_CANCEL) return 1;
    if(answer==MBOX_CLICKED_YES){
      if(!namechanged){
        FXFileDialog savedialog(this,"Save Document");
        savedialog.setFilename(filename);
        if(!savedialog.execute()) return 1;
        if(fxexists(savedialog.getFilename().text())){
          if(MBOX_CLICKED_NO==FXMessageBox::question(this,MBOX_YES_NO,"Overwrite Document","Overwrite existing document?")) return 1;
          }
        filename=savedialog.getFilename();
        }
      saveFile(filename);
      }
    }
  writeRegistry();
  getApp()->exit(0);
  return 1;
  }


// Update title
long TextWindow::onUpdTitle(FXObject* sender,FXSelector,void*){
  FXString title="FOX Text Editor:- " + filename;
  if(editor->isModified()) title+="*";
  sender->handle(this,MKUINT(FXWindow::ID_SETSTRINGVALUE,SEL_COMMAND),(void*)&title);
  return 1;
  }


// Print the text
long TextWindow::onCmdPrint(FXObject*,FXSelector,void*){
  FXPrintDialog dlg(this,"Print File");
  FXPrinter printer;
  if(dlg.execute()){
    dlg.getPrinter(printer);
    FXTRACE((100,"Printer = %s\n",printer.name));
    }
  return 1;
  }


// Command from the tree list
long TextWindow::onCmdTreeList(FXObject*,FXSelector,void* ptr){
  FXTreeItem *item=(FXTreeItem*)ptr;
  FXuint answer;
  if(!item || !dirlist->isItemFile(item)) return 1;
  if(editor->isModified()){
    answer=FXMessageBox::question(this,MBOX_YES_NO_CANCEL,"Unsaved Document","Save current document to file?");
    if(answer==MBOX_CLICKED_CANCEL) return 1;
    if(answer==MBOX_CLICKED_YES){
      if(!namechanged){
        FXFileDialog savedialog(this,"Save Document");
        if(!savedialog.execute()) return 1;
        if(fxexists(savedialog.getFilename().text())){
          if(MBOX_CLICKED_NO==FXMessageBox::question(this,MBOX_YES_NO,"Overwrite Document","Overwrite existing document?")) return 1;
          }
        filename=savedialog.getFilename();
        }
      saveFile(filename);
      }
    }
  filename=dirlist->getItemPathname(item);
  mrufiles.appendFile(filename);
  loadFile(filename);
  namechanged=TRUE;
  return 1;
  }


// See if we can get it as a filename
long TextWindow::onEditDNDDrop(FXObject*,FXSelector,void*){
  FXuchar *data; FXuint len;
  if(getDNDData(FROM_DRAGNDROP,urilistType,data,len)){
    FXRESIZE(&data,FXuchar,len+1); data[len]='\0';
    FXString urilist=(const FXchar *)data;
    FXString file=FXURL::fileFromURL(urilist.before('\r'));
    FXTRACE((100,"urilist=%s file=%s\n",data,file.text()));
    FXFREE(&data);
    if(file.empty()) return 1;
    filename=file;
    dirlist->setCurrentFile(filename);
    mrufiles.appendFile(filename);
    loadFile(filename);
    namechanged=TRUE;
    return 1;
    }
  return 0;
  }


// See if a filename is being dragged over the window
long TextWindow::onEditDNDMotion(FXObject*,FXSelector,void*){
  if(!editor->isModified()){
    if(offeredDNDType(FROM_DRAGNDROP,urilistType)){
      acceptDrop(DRAG_COPY);
      return 1;
      }
    }
  return 0;
  }


// Change both text and tree background color
long TextWindow::onCmdBackColor(FXObject*,FXSelector,void* ptr){
  editor->setBackColor((FXColor)(long)ptr);
  dirlist->setBackColor((FXColor)(long)ptr);
  return 1;
  }


// Update background color
long TextWindow::onUpdBackColor(FXObject* sender,FXSelector,void* ptr){
  editor->handle(sender,MKUINT(FXText::ID_BACK_COLOR,SEL_UPDATE),ptr);
  return 1;
  }

// Change both text and tree selected background color
long TextWindow::onCmdSelBackColor(FXObject*,FXSelector,void* ptr){
  editor->setSelBackColor((FXColor)(long)ptr);
  dirlist->setSelBackColor((FXColor)(long)ptr);
  return 1;
  }


// Update selected background color
long TextWindow::onUpdSelBackColor(FXObject* sender,FXSelector,void* ptr){
  editor->handle(sender,MKUINT(FXText::ID_SELBACK_COLOR,SEL_UPDATE),ptr);
  return 1;
  }


// Change both text and tree text color
long TextWindow::onCmdTextColor(FXObject*,FXSelector,void* ptr){
  editor->setTextColor((FXColor)(long)ptr);
  dirlist->setTextColor((FXColor)(long)ptr);
  return 1;
  }


// Forward GUI update to text widget
long TextWindow::onUpdTextColor(FXObject* sender,FXSelector,void* ptr){
  editor->handle(sender,MKUINT(FXText::ID_TEXT_COLOR,SEL_UPDATE),ptr);
  return 1;
  }


// Change both text and tree
long TextWindow::onCmdSelTextColor(FXObject*,FXSelector,void* ptr){
  editor->setSelTextColor((FXColor)(long)ptr);
  dirlist->setSelTextColor((FXColor)(long)ptr);
  return 1;
  }


// Forward GUI update to text widget
long TextWindow::onUpdSelTextColor(FXObject* sender,FXSelector,void* ptr){
  editor->handle(sender,MKUINT(FXText::ID_SELTEXT_COLOR,SEL_UPDATE),ptr);
  return 1;
  }


// Change both text and tree
long TextWindow::onCmdLineColor(FXObject*,FXSelector,void* ptr){
  dirlist->setLineColor((FXColor)(long)ptr);
  return 1;
  }


// Forward GUI update to text widget
long TextWindow::onUpdLineColor(FXObject* sender,FXSelector,void*){
  FXColor clr=dirlist->getLineColor();
  sender->handle(sender,MKUINT(FXWindow::ID_SETINTVALUE,SEL_COMMAND),(void*)&clr);
  return 1;
  }


// Text inserted
long TextWindow::onTextInserted(FXObject*,FXSelector,void* ptr){
  FXint *what=(FXint*)ptr;
  FXTRACE((140,"Inserted: %d %d\n",what[0],what[1]));
  undolist.add(new FXTextInsert(editor,what[0],what[1]));
  
  // Keep the undo list in check by trimming it down to KEEPUNDOSIZE
  // whenever the amount of undo buffering exceeds MAXUNDOSIZE.
  if(undolist.undoSize()>MAXUNDOSIZE) undolist.trimSize(KEEPUNDOSIZE);
  return 1;
  }


// Text deleted
long TextWindow::onTextDeleted(FXObject*,FXSelector,void* ptr){
  FXint *what=(FXint*)ptr;
  FXTRACE((140,"Deleted: %d %d\n",what[0],what[1]));
  undolist.add(new FXTextDelete(editor,what[0],what[1]));
  
  // Keep the undo list in check by trimming it down to KEEPUNDOSIZE
  // whenever the amount of undo buffering exceeds MAXUNDOSIZE.
  if(undolist.undoSize()>MAXUNDOSIZE) undolist.trimSize(KEEPUNDOSIZE);
  return 1;
  }


// Search text
long TextWindow::onCmdSearch(FXObject*,FXSelector,void*){
  FXSearchDialog search(this,"Search Text");
  search.setSearchDir(TRUE);
  search.setSearchText("Jeroen");
  search.setSearchMode(SEARCHMODE_IGNORECASE);
  search.execute();
  return 1;
  }


// Replace text
long TextWindow::onCmdReplace(FXObject*,FXSelector,void*){
  FXReplaceDialog replace(this,"Replace Text");
  replace.execute();
  return 1;
  }


// Start the ball rolling
void TextWindow::start(int& argc,char** argv){
  if(argc>1){
    filename=argv[1];
    dirlist->setCurrentFile(filename);
    loadFile(filename);
    namechanged=TRUE;
    }
  }


// Create and show window
void TextWindow::create(){
  if(!urilistType){urilistType=getApp()->registerDragType(urilistTypeName);}
  readRegistry();
  FXMainWindow::create();
  show();
  }


/*******************************************************************************/


// Start the whole thing
int main(int argc,char *argv[]){

  // Make application
  FXApp application("TextEdit","FoxTest");
  
  // Open display
  application.init(argc,argv);

  // Make window
  TextWindow* window=new TextWindow(&application);
  
  // Handle interrupt to save stuff nicely
  application.addSignal(SIGINT,window,TextWindow::ID_QUIT);

  // Create it
  application.create();
  
  // Start
  window->start(argc,argv);

  // Run
  return application.run();
  }


